//    nuni.c    A boot loader for linux which does not use the bios.
//              For ext2 file system and IDE drives.
//    Copyright (C) 1999-2000  Neil Koozer
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//    Neil Koozer, 907 SE Sharon Ave, Roseburg, OR 97470-3824
//    nkoozer@rosenet.net
//-------------------------------------------------------------------

#define verbose 0    //1=verbose 0=quiet
#define ontrack 0    //set to 63 if Ontrack Disk Manager is on drive

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <linux/hdreg.h>

struct stat st, st2;
struct hd_geometry geo;
char dev_name[11] = "/dev/hdx";

unsigned int load_address, iblock, ioffset, iobase, iobase2, drive;
unsigned int total_blocks, blocks_left, move_count, move_start;
unsigned int block_bytes, itable;
unsigned char ch, major, minor, setup_sects, setup_blocks, double_blks,
block_size;

char superblock[128];
unsigned int* blk_size  = (unsigned int*)&superblock[24];
unsigned int* frag_size = (unsigned int*)&superblock[28];
unsigned int* ipg       = (unsigned int*)&superblock[40];
unsigned int* revision  = (unsigned int*)&superblock[76];

char number[3];
FILE* outfile;
int temp, fd, doubles;

int get_inode(int inode){
    int temp;
    temp = inode / *ipg;	//temp = group containing this inode
    if (*blk_size == 0 || *blk_size == 1)
        temp = temp*32+0x808;	//temp = offset in group table to
    else if (*blk_size == 2)	//find block# of inode table
        temp = temp*32+0x1008;
    else {
        printf( "ext2 block size %u not supported\n",block_bytes);
        return(-1); }

    if (lseek( fd,temp,SEEK_SET) != temp) {
        printf( "lseek1 %s: %s\n",dev_name,strerror(errno)); return(-1); }
    if (read( fd, &itable, 4) !=4) {
        printf( "read1 %s: %s\n",dev_name,strerror(errno)); return(-1); }
	//itable = block# of start of inode table

	//derive block# containing inode:
    iblock = itable + (inode%*ipg)/(4*block_size);

	//now get byte offset of inode in block
    ioffset = ((inode%*ipg)%(4*block_size))*128;
    }

main( int argc, char *argv[] ){

if ( argc != 2 ) {
    printf("Usage:\nnuni pathname\n\n"); return(-1); }

	//get device and inode for chosen file

if (stat( argv[1], &st) != 0) {
    printf( "stat %s: %s\n",argv[1],strerror(errno)); return(-1); }
major = st.st_dev >> 8;
minor = st.st_dev & 0xFF;    

	//Get the number of setup sectors

if ((fd = open( argv[1], O_RDWR )) < 0) {
    printf( "open4 %s: %s\n",argv[1],strerror(errno)); return(-1); }
if (lseek( fd,497,SEEK_SET) != 497) {
    printf( "lseek2 %s: %s\n",argv[1],strerror(errno)); return(-1); }
if (read( fd, &setup_sects, 1) !=1) {
    printf( "read2 %s: %s\n",argv[1],strerror(errno)); return(-1); }
setup_sects ++;

	//Get kernel size (zImage or bzImage)

if (lseek( fd,529,SEEK_SET) != 529) {
    printf( "lseek3 %s: %s\n",argv[1],strerror(errno)); return(-1); }
if (read( fd, &ch, 1) !=1) {
    printf( "read3 %s: %s\n",argv[1],strerror(errno)); return(-1); }
if (ch == 0) load_address = 0x10000;
else load_address = 0x100000;
if (fd) close(fd);

 	//derive device name and select IDE ports

switch (major) {
    case  3: dev_name[7]='a'; iobase=0x1f0; iobase2=0x3f6; break;
    case 22: dev_name[7]='c'; iobase=0x170; iobase2=0x376; break;
    case 33: dev_name[7]='e'; iobase=0x6b00; iobase2=0x6f00; break;
    case 34: dev_name[7]='g'; iobase=0x7300; iobase2=0x7700; break;
    default: printf( "device %d:%d not recognized\n",major,minor); return(-1);
    }
drive = 0x40000000;		//4 means LBA
if (minor > 63) {dev_name[7]++; minor -= 64; drive += 0x10000000;}
sprintf( number, "%u", minor );
strcat( dev_name, number );    

	//get sector address (geo.start) of named partition
    
if ((fd = open( dev_name, O_RDONLY )) < 0) {
    printf( "open %s: %s\n",dev_name,strerror(errno)); return(-1); }
if (ioctl( fd,HDIO_GETGEO,&geo) < 0) {
    printf( "HDIO_GETGEO %s: %s\n",dev_name,strerror(errno)); return(-1); }

	//get superblock for blocksize and ipg (inodes per block)

if (lseek( fd,0x400,SEEK_SET) != 0x400) {
    printf( "lseek4 %s: %s\n",dev_name,strerror(errno)); return(-1); }
if (read( fd, superblock, 128) !=128) {
    printf( "read4 %s: %s\n",dev_name,strerror(errno)); return(-1); }
	//mess with block size
if (*blk_size != *frag_size) {
    printf("block size and frag size are different\n"); return(-1); }
block_size = 2 << ( *blk_size );
block_bytes = 512 * block_size;

if (get_inode(st.st_ino-1)== -1) return(-1);

setup_blocks = (setup_sects + block_size -1) / block_size;
move_count = setup_blocks * block_bytes - setup_sects * 512;
move_start = 0x90000 + setup_sects * 512;

total_blocks = (st.st_size + block_bytes -1) / block_bytes;
blocks_left = total_blocks - setup_blocks;
doubles = total_blocks - block_bytes/4 -12;   //blocks of double indirect
if (doubles < 1) double_blks = 0;
else double_blks = (doubles + block_bytes/4 -1) / (block_bytes/4);


	//make an "include" file for the assembler

if ((outfile = fopen( "nuni.dat", "w" )) == NULL) {
    printf( "fopen %s: %s\n","nuni.dat",strerror(errno)); return(-1); }
fprintf(outfile,"iobase  equ 0x%x\n",iobase);
fprintf(outfile,"iobase2 equ 0x%x\n",iobase2);
fprintf(outfile,"drive   equ 0x%x\n",drive);
fprintf(outfile,"Start   equ %u\n",geo.start + ontrack);
fprintf(outfile,"iblock  equ %u\n",iblock);
fprintf(outfile,"ioffset equ %u\n",ioffset);
fprintf(outfile,"load_address equ 0x%x\n",load_address);
fprintf(outfile,"block_size   equ %u\n",block_size);
fprintf(outfile,"blk_shift    equ %u\n",*blk_size + 1);
fprintf(outfile,"double_blks  equ %u\n",double_blks);
fprintf(outfile,"setup_blocks equ %u\n",setup_blocks);
fprintf(outfile,"blocks_left  equ %u\n",blocks_left);
fprintf(outfile,"move_count   equ 0x%x\n",move_count/4);
fprintf(outfile,"move_start   equ 0x%x\n",move_start);

if (verbose) {
    printf("heads         %u\n", geo.heads);
    printf("sectors       %u\n", geo.sectors);  
    printf("cylinders     %u\n", geo.cylinders);
    printf("major         %u\n", major);
    printf("minor         %u\n", minor);
    printf("block size    %u\n", block_bytes);
    printf("revision      %u\n", *revision);
    printf("dev_name      %s\n", dev_name);
    printf("inode(ref 0)  %u\n", st.st_ino-1);
    printf("file size     %u\n", st.st_size);
    printf("blocks of setup code = %u\n",setup_blocks);
    printf("doubles       %u\n", double_blks);		
    printf("ipg           %u\n", *ipg);
    printf("block# of inode table = %u\n",itable);
    printf("iblock        %u\n", iblock);
    printf("ioffset       %u\n", ioffset);
    printf("start         %u\n", geo.start);
    }
	//now mess with cmdline file

if (stat( "/boot/nuni.kcl", &st2) != 0) {
    printf( "stat %s: %s\n","/boot/nuni.kcl",strerror(errno)); return(-1); }

if (get_inode(st2.st_ino-1)== -1) return(-1);
if (fd)  close(fd);

fprintf(outfile,"iblock2      equ %u\n",iblock);
fprintf(outfile,"ioffset2     equ %u\n",ioffset);
fprintf(outfile,"cmd_buff     equ 0x%x\n",block_bytes*setup_blocks);
fprintf(outfile,"cmd_line     equ 0x%x\n",block_bytes*setup_blocks+st2.st_size);
fclose(outfile);

if (verbose){
    printf("inode2(ref 0) %u\n", st2.st_ino-1);
    printf("cmdfile size  %u\n", st2.st_size);
    printf("iblock        %u\n", iblock);
    printf("ioffset       %u\n", ioffset);
    }

}

